import json
import os

from .multyvac import (
    MultyvacError,
    MultyvacModule,
)

class ConfigError(MultyvacError):
    """Raised when Multyvac's configuration is incomplete or incorrect."""
    pass

class ConfigModule(MultyvacModule):
    """Top-level Config module. Use this through ``multyvac.config``."""
    
    def __init__(self, multyvac, api_key=None, api_secret_key=None,
                 api_url=None):
        MultyvacModule.__init__(self, multyvac)
        
        self.api_key = None
        self.api_secret_key = None
        self.api_url = api_url or 'https://api.multyvac.com'
        
        self._target_uid = None
        self._target_gid = None
        self._detect_sudo()
        
        self._multyvac_path = self.get_multyvac_path()
        self._create_path_ignore_existing(self._multyvac_path)
        self._config_name = 'multyvac.json'
        self._config_path = os.path.join(self._multyvac_path, self._config_name)
        
        self._load_config()
        
        if os.environ.get('MULTYVAC_API_KEY'):
            self.api_key = os.environ.get('MULTYVAC_API_KEY')
            if os.environ.get('MULTYVAC_API_SECRET_KEY'):
                self.api_secret_key = os.environ.get('MULTYVAC_API_SECRET_KEY')
            if os.environ.get('MULTYVAC_API_URL'):
                self.api_url = os.environ.get('MULTYVAC_API_URL')
        
        self.api_key = api_key or self.api_key
        self.api_secret_key = api_secret_key or self.api_secret_key
        self.api_url = api_url or self.api_url or 'https://api.multyvac.com'
    
    def _detect_sudo(self):
        """Determine if sudo is being used. If so, set the target uid/gid so
        that the permissions are set to the calling user."""
        
        if os.name != 'posix':
            return
        
        sudo_uid = os.getenv('SUDO_UID')
        sudo_gid = os.getenv('SUDO_GID', -1)
        sudo_user = os.getenv('SUDO_USER')
        if sudo_uid and sudo_user:
            sudo_user_home = os.path.expanduser('~' + sudo_user)
            home_env = os.getenv('HOME')
            # If "sudo -H" was used, we treat it as if sudo's target user is
            # the true user.
            if home_env == sudo_user_home:
                self._target_uid = int(sudo_uid)
                self._target_gid = int(sudo_gid)
    
    def get_api_domain(self):
        """Returns the domain of the Multyvac API endpoint."""
        return self.api_url[self.api_url.index('//')+2:]
    
    def get_multyvac_path(self):
        """Returns the path on the filesystem to the multyvac configuration
        directory."""
        if os.name == 'nt' and os.environ.get('APPDATA'):
            path = os.path.join(os.environ.get('APPDATA'), 'multyvac')
        else:
            path = os.path.join('~', '.multyvac')
        return os.path.expanduser(path)
    
    def _get_credentials_path(self, api_key=None):
        if api_key:
            return os.path.join(self._multyvac_path,
                                self.get_api_domain(),
                                api_key)
        else:
            return os.path.join(self._multyvac_path,
                                self.get_api_domain())
    
    def _create_path_ignore_existing(self, path):
        """Create a path if it doesn't exist. Ignore any race condition errors
        if the path happens to be created while this is executing."""
        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                # File exists (17) is okay
                if e.errno != 17:
                    raise
            self._fix_permission(path)
    
    def _fix_permission(self, path):
        """If sudo is used, this will attempt to correct the permissions of the
        path to be that of the calling user. However, this may not always be
        possible."""
        if self._target_uid:
            try:
                os.chown(path, self._target_uid, self._target_gid)
            except OSError as e:
                # No such file (2) is okay
                # Permission denied (13) is okay
                if e.error not in (2, 13):
                    raise
    
    def get_auth(self):
        """If a key is set, returns a tuple of (api key, api secret key).
        Otherwise, raises a ConfigError."""
        if not self.api_key:
            raise ConfigError('Api Key is not set')
        return (self.api_key, self.api_secret_key)
        
    def path_to_private_key(self):
        """Returns the path on the file system to the private key necessary to
        SSH into a job. Note that this private key is tied to the currently
        set api key."""
        private_key_path = self._get_credentials_path(self.api_key + '.key')
        if os.path.exists(private_key_path):
            return private_key_path
        else:
            self._get_and_save_ssh_key(self.api_key)
            return private_key_path
    
    def set_key(self, api_key, api_secret_key, api_url=None):
        """
        Sets the API credentials forcibly replacing any other credentials
        that have been set.
        
        :param api_key: The ApiKey id.
        :param api_secret_key: The ApiKey secret.
        :param api_url: The API end point to use.
        """
        self.api_key = api_key
        self.api_secret_key = api_secret_key
        self.api_url = api_url or self.api_url
    
    def _load_config(self):
        if os.path.exists(self._config_path):
            with open(self._config_path) as f:
                conf = json.load(f)
                self.api_key = conf['default_api_key']
                self.api_url = conf['api_url']
                # only load if we have an api key
                if self.api_key:
                    self._load_credential_from_disk(self.api_key)
        else:
            self.save_to_disk()
        
    def _load_credential_from_disk(self, api_key):
        credential_path = self._get_credentials_path(api_key)
        if os.path.exists(credential_path + '.json'):
            with open(credential_path + '.json') as f:
                credential = json.load(f)
            self.api_secret_key = credential['api_secret_key']
        else:
            raise ConfigError('Could not find credentials for api_key '
                                      '%s on disk' % api_key)
        
    def _get_and_save_ssh_key(self, api_key):
        """Assumes that the api_key and secret_key are valid"""
        key = self.multyvac.api_key.get(api_key)
        if not api_key:
            raise KeyError('Could not get find key.')
        
        credentials_path = self._get_credentials_path()
        self._create_path_ignore_existing(credentials_path)
        
        credential_path = self._get_credentials_path(api_key)
        credential_key_path = credential_path + '.key'
        with open(credential_key_path, 'w') as f:
            f.write(key.private_key)
        os.chmod(credential_key_path, 0600)
        self._fix_permission(credential_key_path)
        
    def save_to_disk(self):
        """Saves the current configuration to disk. The next time multyvac
        is imported, the current configurations will be available."""
        with open(self._config_path, 'w') as f:
            d = {'default_api_key': self.api_key,
                 'api_url': self.api_url,
                 }
            json.dump(d, f, indent=4)
        self._fix_permission(self._config_path)
        credentials_path = self._get_credentials_path()
        self._create_path_ignore_existing(credentials_path)
        if self.api_key:
            credential_path = self._get_credentials_path(self.api_key)
            credential_json_path = credential_path + '.json'
            with open(credential_json_path, 'w') as f:
                d = {'api_secret_key': self.api_secret_key}
                json.dump(d, f, indent=4)
            self._fix_permission(credential_json_path)



# Guarantees that setuptools is available
from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup

# We disable requirements.txt parsing for now since users are having problems
# with their cwd being elsewhere than their requirements.txt file.
#from pip.req import parse_requirements
# parse_requirements() returns generator of pip.req.InstallRequirement objects
#install_reqs = [str(ir.req) for ir in parse_requirements('requirements.txt')]

install_reqs = ['requests>=1.1.0', 'ConcurrentLogHandler>=0.9.1']

dist = setup(
    name='multyvac',
    version='0.4.9',
    description='Multyvac for Python',      
    author='Multyvac, Inc.',
    author_email='dev@multyvac.com',
    url='http://www.multyvac.com',
    install_requires=install_reqs,
    license='LICENSE.txt',
    packages=['multyvac', 'multyvac.util'],
    long_description=open('README.rst').read(),
    platforms=['CPython 2.6', 'CPython 2.7'],      
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        ],
)




from .multyvac import (
    Multyvac,
    MultyvacModel,
    MultyvacModule,
)

class ApiKey(MultyvacModel):
    """Represents a Multyvac ApiKey and its associated operations."""
    
    def __init__(self, id, **kwargs):
        """Creates a new ApiKey."""
        MultyvacModel.__init__(self, **kwargs)
        
        self.id = id
        self.secret_key = kwargs.get('secret_key')
        self.public_key = kwargs.get('public_key')
        self.private_key = kwargs.get('private_key')
        self.created = kwargs.get('created')
        self.active = kwargs.get('active')

    def activate(self):
        """Activates the ApiKey."""
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/key/%s/activate' % self.id)
        self.active = True
        return MultyvacModule.check_success(r)

    def deactivate(self):
        """Deactivates the ApiKey."""
        r = self.multyvac._ask(Multyvac._ASK_POST,
                               '/key/%s/deactivate' % self.id)
        self.active = False
        return MultyvacModule.check_success(r)

    def __repr__(self):
        return 'ApiKey(%s)' % repr(self.id)

class ApiKeyModule(MultyvacModule):
    """Top-level ApiKey module. Use this through ``multyvac.api_key``."""
    
    def get(self, id, username=None, password=None):
        """
        Returns an ApiKey.
        
        :param id: The id of the ApiKey to return.
        :param username: Web account credential optionally used in lieu of
            API credentials.
        :param password: Web account credential optionally used in lieu of
            API credentials.
        """
        auth = ('web-'+username, password) if username else None
        iter_in = MultyvacModule.is_iterable_list(id)
        params = {'id': id}
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/key',
                               auth=auth,
                               params=params)
        keys = [ApiKey(multyvac=self.multyvac, **k) for k in r['keys']]
        if iter_in:
            return keys
        else:
            return keys[0] if keys else None
        
        return ApiKey(id=id, multyvac=self.multyvac)
    
    def list(self, username=None, password=None):
        """
        Returns a list of ApiKeys.
        
        :param username: Web account credential optionally used in lieu of
            API credentials.
        :param password: Web account credential optionally used in lieu of
            API credentials.
        """
        auth = ('web-'+username, password) if username else None
        r = self.multyvac._ask(Multyvac._ASK_GET,
                               '/key',
                               auth=auth)
        return [ApiKey(multyvac=self.multyvac, **k) for k in r['keys']]
